use std::fs;

pub fn solve() {
    let path = "res/05/input.txt";

    let contents = fs::read_to_string(path).expect("I/O error, wrong path?");

    let (stacks_str, moves_str) = contents.split_at(contents.find("\n\n").unwrap());

    // Parse stacks
    let stacks_lines: Vec<&str> = stacks_str.split("\n").collect();
    let stacks_max_height = stacks_lines.len() - 1;
    let stacks_num = (stacks_lines[0].len() + 1) / 4;
    let mut stacks: Vec<Stack> = (0..stacks_num).map(|_| Stack::new()).collect();
    (0..stacks_max_height).rev().for_each(|i| {
        for j in 0..stacks_num {
            let chr = stacks_lines[i].chars().collect::<Vec<char>>()[j * 4 + 1];
            if chr != ' ' {
                stacks[j].push(chr);
            }
        }
    });
    // Copy stacks for task 2
    let mut stacks2: Vec<Stack> = stacks
        .iter()
        .map(|stack| Stack {
            stack: stack.stack.clone(),
        })
        .collect();

    // Parse moves
    let moves: Vec<Move> = moves_str
        .split("\n")
        .filter(|line| *line != "")
        .map(|line| {
            let words: Vec<&str> = line.split(" ").collect();
            let amount = words[1].parse::<i32>().unwrap();
            let from = words[3].parse::<i32>().unwrap();
            let to = words[5].parse::<i32>().unwrap();

            Move::new(from, to, amount)
        })
        .collect();

    // Run moves for task 1
    moves.iter().for_each(|mv| {
        for _ in 0..mv.amount {
            let chr = stacks[(mv.from - 1) as usize].pop();
            stacks[(mv.to - 1) as usize].push(chr);
        }
    });

    let res: String = stacks
        .iter()
        .map(|stack| stack.stack.last().unwrap())
        .collect();
    println!("Result 1: {}", res);

    // Run moves for task 2
    moves.iter().for_each(|mv| {
        let mut chrs = Vec::<char>::new();
        for _ in 0..mv.amount {
            chrs.push(stacks2[(mv.from - 1) as usize].pop());
        }
        chrs.iter().rev().for_each(|c| {
            stacks2[(mv.to - 1) as usize].push(*c);
        });
    });

    let res: String = stacks2
        .iter()
        .map(|stack| stack.stack.last().unwrap())
        .collect();
    println!("Result 2: {}", res);
}

struct Stack {
    stack: Vec<char>,
}

impl Stack {
    // stack order: 1st element is at the bottom
    pub fn new() -> Self {
        Stack {
            stack: Vec::<char>::new(),
        }
    }

    pub fn push(&mut self, chr: char) {
        self.stack.push(chr)
    }

    pub fn pop(&mut self) -> char {
        self.stack.pop().unwrap()
    }
}

struct Move {
    pub from: i32,
    pub to: i32,
    pub amount: i32,
}

impl Move {
    pub fn new(from: i32, to: i32, amount: i32) -> Self {
        Move { from, to, amount }
    }
}
