use std::fs;

pub fn solve() {
    let path = "res/08/input.txt";

    let contents = fs::read_to_string(path).expect("I/O error, wrong path?");

    let cont_arr = contents.lines().collect::<Vec<&str>>();

    // parse trees as 2D i32 Vec
    let trees: Vec<Vec<i32>> = cont_arr
        .iter()
        .filter(|line| **line != "")
        .map(|line| {
            line.chars()
                .collect::<Vec<char>>()
                .iter()
                .map(|t| return t.to_digit(10).unwrap() as i32)
                .collect()
        })
        .collect();

    let width = trees[0].len();
    let height = trees.len();

    // task 1
    // initialize visibilities with 0
    let mut visibilities: Vec<Vec<i32>> = (0..height)
        .map(|_| (0..width).map(|_| 0).collect())
        .collect();

    // mark visible trees with 1
    // left to right & right to left
    for i in 0..height {
        let mut last_max_left = -1;
        let mut last_max_right = -1;

        for j in 0..width {
            if trees[i][j] > last_max_left {
                visibilities[i][j] = 1;
                last_max_left = trees[i][j];
            }
            if trees[i][width - j - 1] > last_max_right {
                visibilities[i][width - j - 1] = 1;
                last_max_right = trees[i][width - j - 1];
            }
        }
    }

    // top to bot & bot to top
    for i in 0..width {
        let mut last_max_top = -1;
        let mut last_max_bot = -1;

        for j in 0..height {
            if trees[j][i] > last_max_top {
                visibilities[j][i] = 1;
                last_max_top = trees[j][i];
            }
            if trees[width - j - 1][i] > last_max_bot {
                visibilities[height - j - 1][i] = 1;
                last_max_bot = trees[height - j - 1][i];
            }
        }
    }

    // sum all visible trees
    let res = visibilities
        .iter()
        .map(|line| line.iter().sum::<i32>())
        .sum::<i32>();

    println!("Result 1: {}", res);

    // task 2
    let scenic_scores: Vec<Vec<i32>> = (0..height)
        .map(|i| {
            (0..width)
                .map(|j| {
                    // calculate view range in each direction
                    let mut to_top = 0;
                    for k in (0..i).rev() {
                        to_top += 1;
                        if trees[k][j] >= trees[i][j] {
                            break;
                        }
                    }

                    let mut to_bot = 0;
                    for k in i + 1..height {
                        to_bot += 1;
                        if trees[k][j] >= trees[i][j] {
                            break;
                        }
                    }

                    let mut to_left = 0;
                    for k in (0..j).rev() {
                        to_left += 1;
                        if trees[i][k] >= trees[i][j] {
                            break;
                        }
                    }

                    let mut to_right = 0;
                    for k in j + 1..width {
                        to_right += 1;
                        if trees[i][k] >= trees[i][j] {
                            break;
                        }
                    }

                    to_top * to_bot * to_left * to_right
                })
                .collect()
        })
        .collect();

    // find max scenic score
    let res = scenic_scores
        .iter()
        .map(|line| line.iter().max().unwrap())
        .max()
        .unwrap();

    println!("Result 2: {}", res)
}
