use std::fs;

const CAVE_WIDTH: u32 = 400;
const CAVE_HEIGHT: u32 = 200;
const X_OFFSET: u32 = 300;
const START_POINT: (u32, u32) = (500, 0);

pub fn solve() {
    let path = "res/14/input.txt";

    let contents = fs::read_to_string(path).expect("I/O error, wrong path?");

    // parse paths
    let rock_paths: Vec<Vec<(u32, u32)>> = contents
        .lines()
        .filter(|line| *line != "")
        .map(|line| {
            line.split(" -> ")
                .map(|coord| {
                    let coord_arr: Vec<&str> = coord.split(",").collect();
                    (
                        coord_arr[0].parse::<u32>().unwrap(),
                        coord_arr[1].parse::<u32>().unwrap(),
                    )
                })
                .collect()
        })
        .collect();

    // cave[X][Y]
    let mut cave: Vec<Vec<bool>> = (0..CAVE_WIDTH)
        .map(|_| (0..CAVE_HEIGHT).map(|_| false).collect())
        .collect();

    // set cave path structure
    rock_paths.iter().for_each(|path| {
        path.windows(2).for_each(|pairs| {
            // draw rock lines
            let x_iter = if pairs[0].0 > pairs[1].0 {
                pairs[1].0..=pairs[0].0
            } else {
                pairs[0].0..=pairs[1].0
            };
            x_iter.for_each(|x| {
                let y_iter = if pairs[0].1 > pairs[1].1 {
                    pairs[1].1..=pairs[0].1
                } else {
                    pairs[0].1..=pairs[1].1
                };
                y_iter.for_each(|y| {
                    cave[(x - X_OFFSET) as usize][y as usize] = true;
                })
            });
        })
    });

    // task 1 - normal cave
    // let sand fall
    let mut sand_unit_counter: u32 = 0;
    let mut sand_pos = next_sand_pos(&cave);
    while sand_pos.1 < CAVE_HEIGHT - 1 {
        cave[sand_pos.0 as usize][sand_pos.1 as usize] = true;
        sand_pos = next_sand_pos(&cave);
        sand_unit_counter += 1;
    }

    println!("Result 1: {sand_unit_counter}");

    // task 2 - cave with ground
    // find ground - cave height is 2 + max y value
    let cave_height = 2 + rock_paths
        .iter()
        .map(|line| line.iter().map(|r| r.1).max().unwrap())
        .max()
        .unwrap();
    // build ground
    cave.iter_mut()
        .for_each(|col| col[(cave_height) as usize] = true);

    // let sand fall again, until it reaches the start point
    while sand_pos != (START_POINT.0 - X_OFFSET, START_POINT.1) {
        cave[sand_pos.0 as usize][sand_pos.1 as usize] = true;
        sand_pos = next_sand_pos(&cave);
        sand_unit_counter += 1;
    }

    println!("Result 2: {sand_unit_counter}");
}

// returns the next position a sand unit will come to rest at
fn next_sand_pos(cave: &Vec<Vec<bool>>) -> (u32, u32) {
    let start_pos = (START_POINT.0 - X_OFFSET, START_POINT.1);
    let mut pos: (u32, u32) = start_pos;
    while !is_end_pos(pos, cave) {
        // first check directly below
        if !cave[(pos.0) as usize][(pos.1 + 1) as usize] {
            pos = (pos.0, pos.1 + 1)
        }
        // then diagonally down-left
        else if !cave[(pos.0 - 1) as usize][(pos.1 + 1) as usize] {
            pos = (pos.0 - 1, pos.1 + 1)
        }
        // then diagonally down-right
        else if !cave[(pos.0 + 1) as usize][(pos.1 + 1) as usize] {
            pos = (pos.0 + 1, pos.1 + 1)
        }
    }

    pos
}

// returns whether the passed coordinates are a valid end position for a sand unit
fn is_end_pos(pos: (u32, u32), cave: &Vec<Vec<bool>>) -> bool {
    if pos.1 as usize == cave[0].len() - 1 {
        // pos is on ground of the cave, lays still
        return true;
    }
    // don't need to check for enough space to left+right since cave is big enough
    if cave[(pos.0 - 1) as usize][(pos.1 + 1) as usize]
        && cave[(pos.0) as usize][(pos.1 + 1) as usize]
        && cave[(pos.0 + 1) as usize][(pos.1 + 1) as usize]
    {
        return true;
    }

    false
}

// prints the cave to stdout
fn _print_cave(cave: &Vec<Vec<bool>>) {
    let mut out: Vec<String> = (0..CAVE_HEIGHT).map(|_| "".to_string()).collect();
    cave.iter().for_each(|line| {
        line.iter().enumerate().for_each(|(i, cell)| match *cell {
            true => out[i] += "#",
            false => out[i] += ".",
        });
    });
    out.iter().for_each(|s| println!("{s}"));
}
