use std::{collections::HashSet, fs};

const ROW: i32 = 2_000_000;
const TASK_2_CONST: i32 = 4_000_000;

pub fn solve() {
    let path = "res/15/input.txt";

    let contents = fs::read_to_string(path).expect("File read error");

    // parse sensors as tuples with (sensor, beacon, radius)
    let sensors: Vec<((i32, i32), (i32, i32), i32)> = contents
        .lines()
        .filter(|line| *line != "")
        .map(|line| {
            let line = line
                .replace("x", "")
                .replace("y", "")
                .replace("=", "")
                .replace(":", "")
                .replace(",", "");

            let words: Vec<&str> = line.split_whitespace().collect();

            let x1 = words[2].parse::<i32>().unwrap();
            let y1 = words[3].parse::<i32>().unwrap();
            let x2 = words[8].parse::<i32>().unwrap();
            let y2 = words[9].parse::<i32>().unwrap();
            let rad = x1.abs_diff(x2) as i32 + y1.abs_diff(y2) as i32;
            ((x1, y1), (x2, y2), rad)
        })
        .collect();

    // task 1
    // go through sensors, add x-indices to hashset which are covered by sensors on y=2_000_000
    let mut row_map = HashSet::<i32>::new();
    sensors.iter().for_each(|sensor| {
        let ((x1, y1), (_, _), rad) = sensor;

        let y_diff = y1.abs_diff(ROW) as i32;
        let x_diff = rad - y_diff;
        if x_diff < 0 {
            return;
        }

        (x1 - x_diff..=x1 + x_diff).for_each(|i| {
            row_map.insert(i);
        })
    });
    // remove all x-indices with beacons
    sensors.iter().for_each(|sensor| {
        let ((_, _), (x2, y2), _) = sensor;

        if *y2 == ROW && row_map.contains(&x2) {
            row_map.remove(&x2);
        }
    });

    let res = row_map.len();
    println!("Result 1: {res}");

    // task 2
    // build vec with all outline points of sensor ranges
    let mut outline = Vec::<(i32, i32)>::new();
    sensors.iter().for_each(|sensor| {
        let ((x, y), (_, _), rad) = *sensor;
        let mut curr_x = x + rad + 1;
        let mut curr_y = y;
        for _ in 0..=rad {
            curr_x -= 1;
            curr_y += 1;
            outline.push((curr_x, curr_y));
        }
        for _ in 0..=rad {
            curr_x -= 1;
            curr_y -= 1;
            outline.push((curr_x, curr_y));
        }
        for _ in 0..=rad {
            curr_x += 1;
            curr_y -= 1;
            outline.push((curr_x, curr_y));
        }
        for _ in 0..=rad {
            curr_x += 1;
            curr_y += 1;
            outline.push((curr_x, curr_y));
        }
    });

    // find point that is out of RAD for each sensor
    let mut res_point = (-1, -1);
    for p in outline {
        let mut point_in_sensor_range = false;
        for s in &sensors {
            let dist = p.0.abs_diff(s.0 .0) as i32 + p.1.abs_diff(s.0 .1) as i32;
            if dist <= s.2 {
                point_in_sensor_range = true;
                break;
            }
        }
        if !point_in_sensor_range
            && p.0 >= 0
            && p.0 <= TASK_2_CONST
            && p.1 >= 0
            && p.1 <= TASK_2_CONST
        {
            res_point = p;
            break;
        }
    }

    let res = TASK_2_CONST as i64 * res_point.0 as i64 + res_point.1 as i64;
    println!("Result 2: {res}");
}
